#ifndef SYNC_DATA_H_
#define SYNC_DATA_H_

#include <deque>
#include <map>
#include <unordered_map>
//#include <stdlib.h>
#include <condition_variable>
#include <fstream>
#include <iostream>
#include <mutex>
#include <thread>
#include <type_traits>
#include <vector>

// Eigen
#include <Eigen/Core>
#include <Eigen/Eigen>
// Mixstruct
#include "lane_msg_info.h"
#include "matched_frame.h"
#include "cam_data.h"
#include "map_data.h"
namespace Localization
{
namespace PointLoc
{
constexpr double MAXTIMEDIFF = 0.02;
const std::string pcddir = "/home/wangmeng/Documents/04xx/";

class SyncData {
public:
  SyncData();
  ~SyncData();
  void Init();

  void SetInput(MapSwitched &mapLaneMsg);
  void SetInput(J2Data &camLaneMsg);
  void SetEgoInput(Localization::PointLoc::InsInfo &egoInfo);
  bool GetFrame(MatchFrame &MixData);
  bool GetOutput(MatchFrame &MixData);


private:
  template <typename T>
  static bool comp(const T &a, const double &b) {
    if constexpr (std::is_same<T, MapSwitched>::value ||
                  std::is_same<T, Localization::PointLoc::InsInfo>::value) {
      return a.stamp_ < b;
    } else if constexpr (std::is_same<T, Eigen::Vector3d>::value) {
      return a.x() < b;
    }
  };
  // switch
  void SwitchMapLaneInfo(
      const MapSwitched &mapLaneRaw, Localization::PointLoc::InsInfo &basePose,
      std::vector<Eigen::Vector3d> &mapPtsUTM,
      std::vector<Eigen::Vector3d> &mapPtsVeh,
      std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> &mapLaneswithid,
      std::vector<double> &mapc1s, std::vector<double> &mapc0s);
  void SwitchCamLaneInfoByMap(
      J2Data &camLaneRaw,
      std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> &mapLanes,
      std::vector<double> &mapc1s, std::vector<double> &mapc0s,
      std::vector<Eigen::Vector3d> &mapPtsVeh,
      Localization::PointLoc::InsInfo &basePose,
      std::vector<Eigen::Vector3d> &j2PtsVehDraw,
      std::vector<Eigen::Vector3d> &j2PtsUTMDraw,
      std::vector<Eigen::Vector3d> &mapPtsVehMatch,
      std::vector<Eigen::Vector3d> &mapPtsUTMMatch,
      std::vector<Eigen::Vector3d> &j2PtsVehMatch,
      std::vector<Eigen::Vector3d> &j2PtsUTMMatch,
      std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> &camvaild,
      std::vector<std::pair<int, std::vector<Eigen::Vector3d>>> &mapvaild,
      std::vector<std::pair<int, int>> &matches);
  // sync
  bool GetNearestEgo(double &stamp,Localization::PointLoc::InsInfo &NearestEgo);
  bool GetNearestMap(double &stamp, MapSwitched &NearestMap);
  void SyncThread();
  void Interpolate3DBetweenTwoPointsByMeter(const Eigen::Vector3d &a,
                                            const Eigen::Vector3d &b,
                                            const double &step,
                                            std::vector<Eigen::Vector3d> &res);

  // func
  Eigen::Vector2d PolyFit(Eigen::Vector2d &xvals, Eigen::Vector2d &yvals);
  void Interpolate();

private:
  std::deque<MapSwitched> mapMsgBuffer_;
  std::deque<Localization::PointLoc::InsInfo> egoMsgBuffer_;
  std::deque<J2Data> j2MsgBuffer_;
  std::deque<MatchFrame> synced_;
  J2Data j2Msg_;

  // switch
  bool isInterpolate_ = true;

  // status
  bool mapInited_ = false;
  bool egoInited_ = false;
  bool j2Inited_ = false;

  // mutex
  std::mutex egomtx_;
  std::mutex mapmtx_;
  std::mutex cammtx_;
  std::mutex sycmtx_;
  std::condition_variable con_;
};
} // namespace PointLoc
} // namespace Localization

#endif
